/*
 * lcd2usb.c - based on test application for the lcd2usb interface
 *             http://www.harbaum.org/till/lcd2usb
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <usb.h>

/* vendor and product id */
#define LCD2USB_VID  0x16c0
#define LCD2USB_PID  0x05dc

/* target is a bit map for CMD/DATA */
#define LCD_CTRL_0         (1<<3)
#define LCD_CTRL_1         (1<<4)
#define LCD_BOTH           (LCD_CTRL_0 | LCD_CTRL_1)

#define LCD_ECHO           (0<<5)
#define LCD_CMD            (1<<5)
#define LCD_DATA           (2<<5)
#define LCD_SET            (3<<5)
#define LCD_GET            (4<<5)

/* target is value to set */
#define LCD_SET_CONTRAST   (LCD_SET | (0<<3))
#define LCD_SET_BRIGHTNESS (LCD_SET | (1<<3))
#define LCD_SET_RESERVED0  (LCD_SET | (2<<3))
#define LCD_SET_RESERVED1  (LCD_SET | (3<<3))

/* target is value to get */
#define LCD_GET_FWVER      (LCD_GET | (0<<3))
#define LCD_GET_KEYS       (LCD_GET | (1<<3))
#define LCD_GET_CTRL       (LCD_GET | (2<<3))
#define LCD_GET_RESERVED1  (LCD_GET | (3<<3))

#define SEG_CLEAR 0;
#define SEG_UPUP (1 << 2);
#define SEG_UPLEFT (1 << 1);
#define SEG_UPRIGHT (1 << 6);
#define SEG_MID (1 << 5);
#define SEG_DOWNLEFT (1 << 0);
#define SEG_DOWNRIGHT (1 << 3);
#define SEG_DOWNDOWN (1 << 4);
#define SEG_ALT (1 << 7);


#ifdef WIN
#include <windows.h>
#include <winbase.h>
#define MSLEEP(a) Sleep(a)
#else
#define MSLEEP(a) usleep(a*1000)
#endif

char shuffle = 0;
char pgm = 0;
char repeat = 0;
char repeatone = 0;
char dash = 0;
char dot = 0;
char playsymbol = 0;

char getcharcode(char c);
void showstring(char *str);
char segfb[6];

const char msg[] = 
  "                "
  "The quick brown fox jumps over the lazy dogs back ..."
  "                ";

usb_dev_handle      *handle = NULL;

int lcd_send(int request, int value, int index) {
  if(usb_control_msg(handle, USB_TYPE_VENDOR, request, 
		      value, index, NULL, 0, 1000) < 0) {
    fprintf(stderr, "USB request failed!");
    return -1;
  }
  return 0;
}

/* to increase performance, a little buffer is being used to */
/* collect command bytes of the same type before transmitting them */
#define BUFFER_MAX_CMD 4        /* current protocol supports up to 4 bytes */
int buffer_current_type = -1;   /* nothing in buffer yet */
int buffer_current_fill = 0;    /* -"- */
unsigned char buffer[BUFFER_MAX_CMD];

/* command format:
 * 7 6 5 4 3 2 1 0
 * C C C T T R L L
 *
 * TT = target bit map
 * R = reserved for future use, set to 0
 * LL = number of bytes in transfer - 1
 */

/* flush command queue due to buffer overflow / content */
/* change or due to explicit request */
void lcd_flush(void) {
  int request, value, index;
  
  /* anything to flush? ignore request if not */
  if (buffer_current_type == -1)
    return;
  
  /* build request byte */
  request = buffer_current_type | (buffer_current_fill - 1);
  
  /* fill value and index with buffer contents. endianess should IMHO not */
  /* be a problem, since usb_control_msg() will handle this. */
  value = buffer[0] | (buffer[1] << 8);
  index = buffer[2] | (buffer[3] << 8);
  
  /* send current buffer contents */
  lcd_send(request, value, index);
  
  /* buffer is now free again */
  buffer_current_type = -1;
  buffer_current_fill = 0;
}

/* enqueue a command into the buffer */
void lcd_enqueue(int command_type, int value) {
  if ((buffer_current_type >= 0) && (buffer_current_type != command_type))
    lcd_flush();
  
  /* add new item to buffer */
  buffer_current_type = command_type;
  buffer[buffer_current_fill++] = value;
  
  /* flush buffer if it's full */
  if (buffer_current_fill == BUFFER_MAX_CMD)
    lcd_flush();
}

/* see HD44780 datasheet for a command description */
void lcd_command(const unsigned char ctrl, const unsigned char cmd) {
  lcd_enqueue(LCD_CMD | ctrl, cmd);
}

/* clear display */
void lcd_clear(void) {
  lcd_command(LCD_BOTH, 0x01);    /* clear display */
  lcd_command(LCD_BOTH, 0x03);    /* return home */
}

/* home display */
void lcd_home(void) {
  lcd_command(LCD_BOTH, 0x03);    /* return home */
}

/* write a data string to the first display */
void lcd_writeA(const char *data) {
  int ctrl = LCD_CTRL_0;
  
  while(*data) 
    lcd_enqueue(LCD_DATA | ctrl, *data++);
  
  lcd_flush();
}

void lcd_writeB(const char *data) {
  int ctrl = LCD_CTRL_1;
  
  while(*data) 
    lcd_enqueue(LCD_DATA | ctrl, *data++);
  
  lcd_flush();
}

/* send a number of 16 bit words to the lcd2usb interface */
/* and verify that they are correctly returned by the echo */
/* command. This may be used to check the reliability of */
/* the usb interfacing */
#define ECHO_NUM 100
void lcd_echo(void) {

  int i, nBytes, errors=0;
  unsigned short val, ret;

  for(i=0;i<ECHO_NUM;i++) {
    val = rand() & 0xffff;
    
    nBytes = usb_control_msg(handle, 
	   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
	   LCD_ECHO, val, 0, (char*)&ret, sizeof(ret), 1000);

    if(nBytes < 0) {
      fprintf(stderr, "USB request failed!");
      return;
    }

    if(val != ret)
      errors++;
  }

  if(errors) 
    fprintf(stderr, "ERROR: %d out of %d echo transfers failed!\n", 
	    errors, ECHO_NUM);
  else 
    printf("Echo test successful!\n");
}

/* get a value from the lcd2usb interface */
int lcd_get(unsigned char cmd) {
  unsigned char       buffer[2];
  int                 nBytes;

  /* send control request and accept return value */
  nBytes = usb_control_msg(handle, 
	   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
	   cmd, 0, 0, (char *)buffer, sizeof(buffer), 1000);

  if(nBytes < 0) {
    fprintf(stderr, "USB request failed!");
    return -1;
  }

  return buffer[0] + 256*buffer[1];
}

/* get lcd2usb interface firmware version */
void lcd_get_version(void) {
  int ver = lcd_get(LCD_GET_FWVER);

  if(ver != -1) 
    printf("Firmware version %d.%d\n", ver&0xff, ver>>8);
}

/* get the bit mask of installed LCD controllers (0 = no */
/* lcd found, 1 = single controller display, 3 = dual */
/* controller display */
void lcd_get_controller(void) {
  int ctrl = lcd_get(LCD_GET_CTRL);

  if(ctrl != -1) {
    if(ctrl)
      printf("Installed controllers: %s%s\n", 
	     (ctrl&1)?"CTRL0":"",
	     (ctrl&2)?" CTRL1":"");
    else
      printf("No controllers installed!\n");
  }
}

void writebytes(char c,char a,char b,char f,char d,char e) {
	
	// write bytes
	char buffer[5];
	char tail;
	tail = (a & 1) | ((b & 1) << 1) | ((c & 1) << 2) | (1 << 7) | (playsymbol << 3);
	a = a | 1;
	b = b | 1;
	c = c | 1;
	sprintf(buffer, "%c%c%c%c",a,b,c,tail);
	printf("writeA: 0x%x,0x%x,0x%x,0x%x\n",a,b,c,tail);
	lcd_writeA(buffer);
	tail = (d & 1) | ((e & 1) << 1) | ((f & 1) << 2) | (1 << 7);
	d = d | 1;
	e = e | 1;
	f = f | 1;
	sprintf(buffer, "%c%c%c%c",d,e,f,tail);
	printf("writeB: 0x%x,0x%x,0x%x,0x%x\n",d,e,f,tail);
	lcd_writeB(buffer);
}


void writesegfbfb() {
	writebytes(
		(segfb[0] & 0x0F) | ((segfb[1] & 0x0F) << 4),
		(segfb[2] & 0x0F) | ((segfb[3] & 0x0F) << 4),
		(segfb[4] & 0x0F) | ((segfb[5] & 0x0F) << 4),
			   
		((segfb[0] & 0xF0) >> 4) | (segfb[1] & 0xF0) |(pgm << 3) | (dash << 7),
		((segfb[2] & 0xF0) >> 4) | (segfb[3] & 0xF0) | (shuffle << 3) | (dot << 7),
		((segfb[4] & 0xF0) >> 4) | (segfb[5] & 0xF0) | (repeat << 3) | (repeatone << 7)
	);
}

int main(int argc, char *argv[]) {
  struct usb_bus      *bus;
  struct usb_device   *dev;
  int i;
  char test[5];
  char j;
  char k;
  char l;
  
  usb_init();
  
  usb_find_busses();
  usb_find_devices();
  
  for(bus = usb_get_busses(); bus; bus = bus->next) {
    for(dev = bus->devices; dev; dev = dev->next) {
      if((dev->descriptor.idVendor == LCD2USB_VID) && 
	 (dev->descriptor.idProduct == LCD2USB_PID)) {
	
	printf("Found LCD2USB device on bus %s device %s.\n", 
	       bus->dirname, dev->filename);
	
	/* open device */
	if(!(handle = usb_open(dev))) 
	  fprintf(stderr, "Error: Cannot open USB device: %s\n", 
		  usb_strerror());

	break;
      }
    }
  }
  
  if(!handle) {
    fprintf(stderr, "Error: Could not find LCD2USB device\n");

#ifdef WIN
    printf("Press return to quit\n");
    getchar();
#endif

    exit(-1);
  }
  
  printf("%s\n",argv[1]);

  /* make lcd interface return some bytes to */
  /* test transfer reliability */
  lcd_echo();

	char a = argv[2][0];
  
	dot = (a & 2) >> 1;
	shuffle = (a & 4) >> 2;
	dash = (a & 8) >> 3;
	repeat = (a & 16) >> 4;
	repeatone = (a & 32) >> 5;
	playsymbol = (a & 64) >> 6;
	showstring(argv[1]);
	
	
}

void showstring(char *str) {
	char pos = 0;
	while(*str != 0 && pos < 6) {
		/*
	printf("\n%2d\n",pos);
	*/
		segfb[pos] = getcharcode(*str);
		pos++;
		str++;
	}
	writesegfbfb();
}

char getcharcode(char c) {
	char res = 0;
	switch(c) {
		case '0':
			res |= SEG_UPUP;
			res |= SEG_UPLEFT;
			res |= SEG_UPRIGHT;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNRIGHT;
			res |= SEG_DOWNDOWN;
			break;
		case '1':
			res |= SEG_UPRIGHT;
			res |= SEG_DOWNRIGHT;
			break;
		case '2':
			res |= SEG_UPUP;
			res |= SEG_UPRIGHT;
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNDOWN;
			break;
		case '3':
			res |= SEG_UPUP;
			res |= SEG_UPRIGHT;
			res |= SEG_MID;
			res |= SEG_DOWNRIGHT;
			res |= SEG_DOWNDOWN;
			break;
		case '4':
			res |= SEG_UPLEFT;
			res |= SEG_MID;
			res |= SEG_UPRIGHT;
			res |= SEG_DOWNRIGHT;
			break;
		case '5': case 's':
			res |= SEG_UPUP;
			res |= SEG_UPLEFT;
			res |= SEG_MID;
			res |= SEG_DOWNRIGHT;
			res |= SEG_DOWNDOWN;
			break;
		case '6':
			res |= SEG_UPUP;
			res |= SEG_UPLEFT;
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNRIGHT;
			res |= SEG_DOWNDOWN;
			break;
		case '7':
			res |= SEG_UPUP;
			res |= SEG_UPRIGHT;
			res |= SEG_DOWNRIGHT;
			break;
		case '8':
			res |= SEG_UPUP;
			res |= SEG_UPRIGHT;
			res |= SEG_UPLEFT;
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNRIGHT;
			res |= SEG_DOWNDOWN;
			break;
		case '9':
			res |= SEG_UPUP;
			res |= SEG_UPRIGHT;
			res |= SEG_UPLEFT;
			res |= SEG_MID;
			res |= SEG_DOWNRIGHT;
			res |= SEG_DOWNDOWN;
			break;
		case 'r':
			res |= SEG_DOWNLEFT;
			res |= SEG_UPLEFT;
			res |= SEG_UPRIGHT;
			res |= SEG_UPUP;
			break;
		case 'e':
			res |= SEG_UPUP;
			res |= SEG_UPRIGHT;
			res |= SEG_UPLEFT;
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNDOWN;
			break;
		case 'n':
			res |= SEG_UPUP;
			res |= SEG_UPLEFT;
			res |= SEG_UPRIGHT;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNRIGHT;
			break;
		case 'y':
			res |= SEG_UPLEFT;
			res |= SEG_UPRIGHT;
			res |= SEG_MID;
			res |= SEG_DOWNRIGHT;
			res |= SEG_DOWNDOWN;
			break;
		case 't':
			res |= SEG_UPLEFT;
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNDOWN;
			break;
		case '-':
			res = 0;
			break;
		case 'o':
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNRIGHT;
			res |= SEG_DOWNDOWN;
			break;
		case 'p':
			res |= SEG_UPUP;
			res |= SEG_UPRIGHT;
			res |= SEG_UPLEFT;
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			break;
		case 'v':
			res |= SEG_UPLEFT;
			res |= SEG_UPRIGHT;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNRIGHT;
			res |= SEG_DOWNDOWN;
			break;
		case 'l':
			res |= SEG_UPLEFT;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNDOWN;
			break;
		case 'h':
			res |= SEG_UPLEFT;
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNRIGHT;
			break;
		case 'a':
			res |= SEG_UPUP;
			res |= SEG_UPRIGHT;
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNRIGHT;
			res |= SEG_DOWNDOWN;
			break;
		case 'c':
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNDOWN;
			break;
		case 'C':
			res |= SEG_UPUP;
			res |= SEG_UPLEFT;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNDOWN;
			break;
		case 'k':
			res |= SEG_UPLEFT;
			res |= SEG_UPRIGHT;
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNRIGHT;
			break;
		case 'd':
			res |= SEG_UPRIGHT;
			res |= SEG_MID;
			res |= SEG_DOWNLEFT;
			res |= SEG_DOWNRIGHT;
			res |= SEG_DOWNDOWN;
			break;
		case 'm':
			res |= SEG_UPUP;
			res |= SEG_UPLEFT;
			res |= SEG_UPRIGHT;
			res |= SEG_DOWNLEFT;
			break;
		case 'M':
			res |= SEG_UPUP;
			res |= SEG_UPLEFT;
			res |= SEG_UPRIGHT;
			res |= SEG_DOWNRIGHT;
			break;

	}
	return res;
}
